import numpy as np
import tensorflow as tf
#Function to make the monte carlo simulation. To see more please go to the original file in Scripts
def MonteCarlo(net,Xtest,Ytest,M,Wstd,Bstd,force,Derr=0,net_name="Network",custom_objects=None,dtype='float32',
	optimizer=tf.keras.optimizers.SGD(learning_rate=0.1,momentum=0.9),loss=['sparse_categorical_crossentropy'],metrics=['accuracy'],top5=False):
	"""
	Input Parameters:
	net: Name of the network model you want to test (it must be saved in the folder Models)
	Xtest and Ytest: Validation/Testing dataset
	M: Number of samples for the Monte Carlo
	Wstd and Bstd: Weights and Bias error for the simulation. It must be a float between 0-1
	force: String, should be "yes" or "no" when you want to use a Wstd or Bstd different from the used during training i.e. If you trained A-Connect with 50% and you want to test it with an error of 70%
	you must define force="yes"
	Derr: If you want to introduce a deterministic error when you are using BW in the network. Float between 0-1
	net_name = String with the name you want to use to save the simulation results
	custom_objects: Python dictionary with the name of all the custom elements that you used in your model i.e. If you use an A-Connect model with Conv and FC A-Connect custom_objects should be
	custom_objects= {'ConvAConnect':ConvAConnect.ConvAConnect,'AConnect':AConnect.AConnect}
	SRAMsz: Matrix dimension for the static error matrix that you want to generate. It is depend on the dimension of the layer weights
	SRAMBsz: Vector dimension for the static error vector that you want to generate. It is depend on the dimension of the layer weights
	optimizer,loss,metrics: The values that you used during the training
	This function returns the noisy accuracy values and the mean of this values
	"""

	### Script to change the error matrix during inference or introduce the error to the layer weights
	### HOw to
	"""
	Input Parameters
	net: Loaded model
	Wstd: Weights standard deviation for simulation
	Bstd: Bias standard deviation for simulation
	force: When you want to use the training deviation or the simulation deviation
	Derr: Deterministic error
	SRAMsz: ERror matrix size
	SRAMBsz: Error vector size
	This function returns a NoisyNet and the values of Wstd and Bstd used
	"""
	def add_Wnoise(net,Wstd,Bstd,force,Derr,dtype='float32'):
		layers = net.layers #Get the list of layers used in the model
		Nlayers = np.size(layers) #Get the number of layers
		
		for i in range(Nlayers): #Iterate over the number of layers
			if layers[i].count_params() != 0: #If the layer does not have training parameters it is omitted

				if hasattr(layers[i],'kernel') or hasattr(layers[i],'W'):  #Does the layer have weights or kernel?

					Wsz = np.shape(layers[i].weights[0]) #Takes the weights/kernel size
					Bsz = np.shape(layers[i].weights[1]) #Takes the bias size
					MBerr_aux = np.random.randn(Bsz[0])
					if hasattr(layers[i],'strides'): #If the layer have the attribute strides means that it is a convolutional layer
						Merr_aux = np.random.randn(Wsz[0],Wsz[1],Wsz[2],Wsz[3])
					else:
						Merr_aux = np.random.randn(Wsz[0], Wsz[1]) #If the layer does not have strides, it is a FC layer
					
					if hasattr(layers[i], 'Wstd'): #Does the layer have Wstd? if it is true is an A-Connect or DropConnect network
						if(layers[i].Wstd != 0): #IF the value it is different from zero, the layer is working with the algorithm
							if force == "no": #Do you want to take the training or simulation Wstd value?
								Wstd = layers[i].Wstd
							else:
								Wstd = Wstd
						else: #If it is false, it means that is working as a normal FC layer
							Wstd = Wstd
					else:
						Wstd = Wstd #If it is false, is a FC layers
					if hasattr(layers[i], 'Bstd'): #The same logic is applied for Bstd
						if(layers[i].Bstd != 0):
							if force == "no":
								Bstd = layers[i].Bstd
							else:
								Bstd = Bstd
						else:
							Bstd = Bstd
					else:
						Bstd = Bstd
					if hasattr(layers[i],'Werr') or hasattr(layers[i],'Berr') or hasattr(layers[i],'infWerr') or hasattr(layers[i],'infBerr'): #Now if the layer have Werr or Berr is an A-Conenct or DropConnect layer
						#print(i)#					
						Werr = abs(1+Wstd*Merr_aux) #Create the error matrix taking into account the Wstd and Bstd
						Berr = abs(1+Bstd*MBerr_aux)
						if(layers[i].isBin == 'yes'): 
							if(Derr != 0): #Introduce the deterministic error when BW are used
								weights = layers[i].weights[0]
								wp = weights > 0
								wn = weights <= 0 
								wn = wn.numpy()
								wp = wp.numpy()
								Werr = Derr*wn*Werr + Werr*wp
						if hasattr(layers[i], 'Wstd'):
							if(layers[i].Wstd != 0):
								layers[i].infWerr = Werr #Change the inference error matrix
							else:			
								#print(layers[i].Werr)                            
								layers[i].Werr = Werr
						else:
								layers[i].Werr = Werr					
						if hasattr(layers[i], 'Bstd'):
							if(layers[i].Bstd != 0):
								layers[i].infBerr = Berr #Change the inference error matrix
							else:
								layers[i].Berr = Berr
						else:
							layers[i].Berr = Berr
					else:                #if the layer is not A-Conenct or DropCOnnect the error must be introduced to the weights because it is a normal FC or normal Conv layer
						Werr = abs(1+Wstd*Merr_aux) #Error matrices
						Berr = abs(1+Bstd*MBerr_aux)
						weights = layers[i].weights[0]*Werr #Introduce the mismatch to the weights
						bias = layers[i].weights[1]*Berr #Introduce the mismatch to the bias
						local_weights = [weights,bias] #Create the tuple of modified values
						layers[i].set_weights(local_weights) #Update the values of the weights
	
		NoisyNet = tf.keras.Sequential(layers)
		return NoisyNet,Wstd,Bstd
					

	def classify(net,Xtest,Ytest,top5):
		if top5:
			_, accuracy, top5acc = net.evaluate(Xtest,Ytest,verbose=0,use_multiprocessing=True)
			return accuracy, top5acc
		else:        
			_,accuracy = net.evaluate(Xtest,Ytest,verbose=0,use_multiprocessing=True)
			return accuracy	

	def MCsim(net,Xtest,Ytest,M,Wstd,Bstd,force,Derr=Derr,net_name=net_name,custom_objects=custom_objects,dtype=dtype,
	optimizer=optimizer,loss=loss,metrics=metrics,top5=top5):
		acc_noisy = np.zeros((M,1)) #Initilize the variable where im going to save the noisy accuracy
		top5acc_noisy = np.zeros((M,1)) #Initilize the variable where im going to save the noisy accuracy   top5       
		local_net = tf.keras.models.load_model(net,custom_objects = custom_objects) #Load the trained model
		local_net.save_weights(filepath=(net_name+'_weights.h5')) #Save the weights. It is used to optimize the script RAM consumption
		print(local_net.summary()) #Print the network summary
		print('Simulation Nr.\t | \tWstd\t | \tBstd\t | \tAccuracy | \tTop-5 Accuracy\n')
		print('---------------------------------------------------------------------------------------')
	#	global parallel

		for i in range(M): #Iterate over M samples
			[NetNoisy,Wstdn,Bstdn] = add_Wnoise(local_net,Wstd,Bstd,force,Derr,dtype=dtype) #Function that adds the new noisy matrices to the layers
			NetNoisy.compile(optimizer,loss,metrics) #Compile the model. It is necessary to use the model.evaluate
			if top5:       
				acc_noisy[i],top5acc_noisy[i] = classify(NetNoisy, Xtest, Ytest,top5) #Get the accuracy of the network
				top5acc_noisy[i] = 100*top5acc_noisy[i]              
			else:
				acc_noisy[i] = classify(NetNoisy, Xtest, Ytest,top5) #Get the accuracy of the network                        
			acc_noisy[i] = 100*acc_noisy[i]      
			print('\t%i\t | \t%.1f\t | \t%.1f\t | \t%.2f | \t%.2f\n' %(i,Wstd*100,Bstd*100,acc_noisy[i],top5acc_noisy[i]))
			local_net.load_weights(filepath=(net_name+'_weights.h5')) #Takes the original weights value.
	#		return acc_noisy

		#pool = Pool(mp.cpu_count())
		#acc_noisy = pool.map(parallel, range(M))
		#pool.close()
		media = np.median(acc_noisy)
		IQR = np.percentile(acc_noisy,75) - np.percentile(acc_noisy,25)
		stats = [media,IQR]
		print('---------------------------------------------------------------------------------------')
		print('Median: %.2f%%\n' % media)
		print('IQR Accuracy: %.2f%%\n' % IQR)
		print('Min. Accuracy: %.2f%%\n' % np.amin(acc_noisy))
		print('Max. Accuracy: %.2f%%\n'% np.amax(acc_noisy))

		np.savetxt(net_name+'_simerr_'+str(int(100*Wstd))+'_'+str(int(100*Bstd))+'.txt',acc_noisy,fmt="%.2f") #Save the accuracy of M samples in a txt
		if top5: 
			np.savetxt('TOP5_'+net_name+'_simerr_'+str(int(100*Wstd))+'_'+str(int(100*Bstd))+'.txt',top5acc_noisy,fmt="%.2f") #Save the accuracy of M samples in a txt    
		np.savetxt('Stats_'+net_name+'_simerr_'+str(int(100*Wstd))+'_'+str(int(100*Bstd))+'.txt',stats,fmt="%.2f") #Save the median and iqr of M samples in a txt
		return acc_noisy, media	
	return MCsim(net,Xtest,Ytest,M,Wstd,Bstd,force,Derr=Derr,net_name=net_name,custom_objects=custom_objects,dtype=dtype,
		optimizer=optimizer,loss=loss,metrics=metrics,top5=top5)		
	#Function to do inference. You also could have the top-5 accuracy if you passed to the model metrics and then setting top5=True
def classify(net,Xtest,Ytest,top5):
	def classify(net,Xtest,Ytest,top5):
		if top5:
			_, accuracy, top5acc = net.evaluate(Xtest,Ytest,verbose=0,use_multiprocessing=True)
			return accuracy, top5acc
		else:        
			_,accuracy = net.evaluate(Xtest,Ytest,verbose=0,use_multiprocessing=True)
			return accuracy	
	return classify(net,Xtest,Ytest,top5)
	#Function to load the MNIST dataset. THis function could load the standard 28x28 8 or 4 bits dataset, or 11x11 8 or 4 bits dataset.		
def load_ds(imgSize=[28,28], Quant=8):
	(x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()
	if (imgSize != [28,28]):
		x_train, x_test = np.expand_dims(x_train,-1),np.expand_dims(x_test,-1) #Need an extra dimension to apply tf.image.resize
		x_train = tf.image.resize_with_pad(x_train,imgSize[0],imgSize[1], method=tf.image.ResizeMethod.BICUBIC,antialias=True) #This function applies a resize similar to imresize in matlab
		x_test = tf.image.resize_with_pad(x_test,imgSize[0],imgSize[1], method=tf.image.ResizeMethod.BICUBIC,antialias=True)
		x_train, x_test = np.squeeze(x_train,-1),np.squeeze(x_test,-1) #Remove the extra dimension
	if(Quant !=8):
		xlsb = 256/2**Quant 
		x_train = np.floor(np.divide(x_train,xlsb))
		x_test = np.floor(np.divide(x_test,xlsb))
	
	return (x_train,y_train),(x_test,y_test)
	#Function to plot the box chart
def plotBox(data,labels,legends,color,color_fill,path):
	import matplotlib.pyplot as plt
	"""
	Script to plot and save a box chart with custom style.
	"""

	def plotChart(ax,x,color=color,color_fill=color_fill,labels=labels): #Script that plots the box, needed in the script below
		boxprops = dict(linestyle='-', linewidth=3,facecolor=color_fill,color=color)
		whiskerprops = dict(color=color, linewidth=3)
		flierprops = dict(marker='.', markerfacecolor=color, markersize=6,
		              markeredgecolor=color)   
		capprops = dict(color=color) 
		medianprops = dict(color=color)
		b = ax.boxplot(x,notch=False,widths=0.2,labels=labels,patch_artist=True,
		            boxprops=boxprops,whiskerprops=whiskerprops,flierprops=flierprops,
		            capprops=capprops,medianprops=medianprops
		            )
		return b
		
	def plotBox(data,labels,legends,color,color_fill,path): #Top script  to plot the box with custom style
		"""
		HOW TO:
		data: Data that you want to plot, should be a list or a list of list (maximum 3 list) i.e. data= [data1,data2,data3] where data1 = [x1,x2,...], 
		data2=[y1,y2,...], data3=[z1,z2,...]
		labels: Labels for the x-axis. Must have the same dimension as the data that you are going to plot.
		legends: String. Legends for the plot. Must have the same dimension as the data parameter e.g. data = [data1,data2], legends=[legend1,legend2]
		color: Color for the lines. Should be a list of size 3 with RGB Color.
		color_fill: Color for filling the boxes. Should be a list of size 3 with RGB Color.
		path: String. Where you want to save the image and the name of the archive. By default you do not need to indicate a saving format. By default all
		the images are saved in png format."""
		font = {'family':'serif','style':'normal','weight' : 'semibold',
		    'size'   : 12}      
		fig = plt.figure(figsize=(5,10))
		ax = fig.add_axes([0.1,0.1,0.8,0.8])
		#plt.rcParams['mathtext.rm'] = 'Arial'
		plt.rc('font',**font)            
		ax.set_xlabel("Simulation Error (%)",fontdict={'family':'serif','style':'normal','weight' : 'semibold',
		    'size'   : 15})
		ax.set_ylabel("Validation Accuracy (%)",fontdict={'family':'serif','style':'normal','weight' : 'semibold',
		    'size'   : 17})  
		d_size = len(data)   
		if d_size == 4:     
		    b1 = plotChart(ax,data,color=color,color_fill=color_fill,labels=labels)
		    ax.legend([b1["boxes"][0]],legends, loc='lower left',prop={'family':'serif','style':'normal','weight' : 'semibold',
		    'size'   : 12})     
		elif d_size == 2:
		    b1 = plotChart(ax,data[0],color=color[0],color_fill=color_fill[0],labels=labels)            
		    b2 = plotChart(ax,data[1],color=color[1],color_fill=color_fill[1],labels=labels)
		    ax.legend([b1["boxes"][0], b2["boxes"][0]],[legends[0], legends[1]], loc='lower left',prop={'family':'serif','style':'normal','weight' : 'semibold',
		    'size'   : 12})
		elif d_size == 3:
		    b1 = plotChart(ax,data[0],color=color[0],color_fill=color_fill[0],labels=labels)            
		    b2 = plotChart(ax,data[1],color=color[1],color_fill=color_fill[1],labels=labels)
		    b3 = plotChart(ax,data[2],color=color[2],color_fill=color_fill[2],labels=labels)        
		    ax.legend([b1["boxes"][0], b2["boxes"][0],b3["boxes"][0]],[legends[0], legends[1], legends[2]], loc='lower left',prop={'family':'serif','style':'normal','weight' : 'semibold',
		    'size'   : 12})   
		else:
		    print("Not supported size")         
	 
		ax.spines['top'].set_linewidth(1.4)
		ax.spines['right'].set_linewidth(1.4)
		ax.spines['bottom'].set_linewidth(1.4)
		ax.spines['left'].set_linewidth(1.4)
		plt.savefig(path, bbox_inches='tight')   
	return plotBox(data,labels,legends,color,color_fill,path)



import tensorflow as tf
import numpy as np
############ This layer was made using the template provided by Keras. For more info, go to the official site.
"""
Fully Connected layer with A-Connect
INPUT ARGUMENTS:
-output_size: output_size is the number of neurons of the layer
-Wstd: Wstd standard deviation of the weights(number between 0-1. By default is 0)
-Bstd: Bstd standard deviation of the bias(number between 0-1. By default is 0)       
-isBin: if the layer will binarize the weights(String yes or no. By default is no)
-pool: Number of error that you want to use
-Slice: If you want to slice the batch in order to reduce the memory usage. Only supports slicing the batch into 2,4 or 8 parts. Any other number will not slice the batch.
-d_type: Data type of the weights and other variables. Default is fp32. Please see tf.dtypes.Dtype
-weights_regularizer: Weights regularizer. Default is None
-bias_regularizer: Bias regularizer. Default is None
"""

class FC_AConnect(tf.keras.layers.Layer): 
	def __init__(self,output_size,Wstd=0,Bstd=0,isBin="no",pool=None,Slice=1,d_type=tf.dtypes.float32,weights_regularizer=None,bias_regularizer=None
    ,**kwargs): #__init__ method is the first method used for an object in python to initialize the ...
		super(FC_AConnect, self).__init__()						 		#...object attributes
		self.output_size = output_size							 		#output_size is the number of neurons of the layer
		self.Wstd = Wstd										 		#Wstd standard deviation of the weights(number between 0-1. By default is 0)
		self.Bstd = Bstd										 		#Bstd standard deviation of the bias(number between 0-1. By default is 0)       
		self.isBin = isBin                                       		#if the layer will binarize the weights(String yes or no. By default is no)
		self.pool = pool                                                #Number of error that you want to use
		self.Slice = Slice                                              #If you want to slice the batch in order to reduce the memory usage
		self.d_type = d_type                                            #Data type of the weights and other variables. Default is fp32. Please see tf.dtypes.Dtype
		self.weights_regularizer = tf.keras.regularizers.get(weights_regularizer)                  #Weights regularizer. Default is None
		self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)                        #Bias regularizer. Default is None
		self.validate_init()
	def build(self,input_shape):								 #This method is used for initialize the layer variables that depend on input_shape
								                                    #input_shape is automatically computed by tensorflow
		self.W = self.add_weight("W",							
										shape = [int(input_shape[-1]),self.output_size], #Weights matrix 
										initializer = "glorot_uniform",
                                        dtype = self.d_type,
                                        regularizer = self.weights_regularizer,
										trainable=True)

		self.bias = self.add_weight("bias",
										shape = [self.output_size,],					#Bias vector
										initializer = "zeros",
                                        dtype = self.d_type,
                                        regularizer = self.bias_regularizer,
										trainable=True)					
		if(self.Wstd != 0 or self.Bstd != 0): #If the layer will take into account the standard deviation of the weights or the std of the bias or both
			if(self.Bstd != 0):
				self.infBerr = abs(1+tf.random.normal(shape=[self.output_size],stddev=self.Bstd)) #Bias error vector for inference
				self.infBerr = self.infBerr.numpy()  #It is necessary to convert the tensor to a numpy array, because tensors are constant and therefore cannot be changed
													 #This was necessary to change the error matrix/array when Monte Carlo was running.
				
																	
			else:
				self.Berr = tf.constant(1,dtype=self.d_type)
			if(self.Wstd != 0): 
				self.infWerr = abs(1+tf.random.normal(shape=[int(input_shape[-1]),self.output_size],stddev=self.Wstd)) #Weight matrix for inference
				self.infWerr = self.infWerr.numpy()										 
				
				 
			else:
				self.Werr = tf.constant(1,dtype=self.d_type)
		else:
			self.Werr = tf.constant(1,dtype=self.d_type) #We need to define the number 1 as a float32.
			self.Berr = tf.constant(1,dtype=self.d_type)
		super(FC_AConnect, self).build(input_shape)
		
	def call(self, X, training=None): #With call we can define all the operations that the layer do in the forward propagation.
		self.X = tf.cast(X, dtype=self.d_type)
		row = tf.shape(self.X)[-1]        
		self.batch_size = tf.shape(self.X)[0] #Numpy arrays and tensors have the number of array/tensor in the first dimension.
											  #i.e. a tensor with this shape [1000,784,128] are 1000 matrix of [784,128].
											  #Then the batch_size of the input data also is the first dimension. 
											  
		#This code will train the network. For inference, please go to the else part
		if(training):	
			if(self.Wstd != 0 or self.Bstd != 0):
				if(self.isBin=="yes"):
					weights = self.sign(self.W)			#Binarize the weights and multiply them element wise with Werr mask
				else:
				    weights = self.W
				if(self.pool is None):                                                                                                     
				    if(self.Slice == 2): #Slice the batch into 2 minibatches of size batch/2
				        miniBatch = tf.cast(self.batch_size/2,dtype=tf.int32)                                                                  
				        Z1 = self.slice_batch(weights,miniBatch,0,row) #Takes a portion from 0:minibatch
				        Z2 = self.slice_batch(weights,miniBatch,1,row) #Takes a portion from minibatch:2*minibatch
				        Z = tf.concat([Z1,Z2],axis=0)
				    elif(self.Slice == 4):
				        miniBatch = tf.cast(self.batch_size/4,dtype=tf.int32) #Slice the batch into 4 minibatches of size batch/4
				        Z = self.slice_batch(weights,miniBatch,0,row) #Takes a portion from 0:minibatch
				        for i in range(3):
				            Z1 = self.slice_batch(weights,miniBatch,i+1,row) #Takes a portion from minibatch:2*minibatch
				            Z = tf.concat([Z,Z1],axis=0)                                                                                                                                                                              
				    elif(self.Slice == 8):
				        miniBatch = tf.cast(self.batch_size/8,dtype=tf.int32) #Slice the batch into 8 minibatches of size batch/8                                              
				        Z = self.slice_batch(miniBatch,0,row) #Takes a portion from 0:minibatch                    
				        for i in range(7):                       
				            Z1 = self.slice_batch(miniBatch,i+1,row) #Takes a portion from minibatch:2*minibatch
				            Z = tf.concat([Z,Z1],axis=0)                      
				    else:                 
				        if(self.Wstd !=0):							
				            #Werr = tf.gather(self.Werr,[loc_id])		#Finally, this line will take only N matrices from the "Pool" of error matrices. Where N is the batch size.          
				            Werr = abs(1+tf.random.normal(shape=[self.batch_size,tf.cast(row,tf.int32),self.output_size],stddev=self.Wstd,dtype=self.d_type))#tf.squeeze(Werr, axis=0)				#This is necessary because gather add an extra dimension. Squeeze remove this dimension.			 
                                                                        #That means, with a weights shape of [784,128] and a batch size of 256. Werr should be a tensor with shape	
                                                                        #[256,784,128], but gather return us a tensor with shape [1,256,784,128], so we remove that 1 with squeeze.
				        else:
				            Werr = self.Werr      
				        memW = tf.multiply(weights,Werr)			         	#Finally we multiply element-wise the error matrix with the weights.
                        
				        if(self.Bstd !=0):								#For the bias is exactly the same situation
				            #Berr = tf.gather(self.Berr, [loc_id])  		 
				            Berr = abs(1+tf.random.normal(shape=[self.batch_size,self.output_size],stddev=self.Bstd,dtype=self.d_type))#tf.squeeze(Berr,axis=0)
				        else:
				            Berr = self.Berr
				        membias = tf.multiply(Berr,self.bias)	
                        
				        Xaux = tf.reshape(self.X, [self.batch_size,1,tf.shape(self.X)[-1]]) #We need this reshape, beacuse the input data is a column vector with
                                                                                            # 2 dimension, e.g. in the first layer using MNIST we will have a vector with
                                                                                            #shape [batchsize,784], and we need to do a matrix multiplication.
                                                                                            #Which means the last dimension of the first matrix and the first dimension of the
                                                                                            #second matrix must be equal. In this case, with a FC layer with 128 neurons we will have this dimensions
                                                                                            #[batchsize,784] for the input and for the weights mask [batchsize,784,128]
                                                                                            #And the function tf.matmul will not recognize the first dimension of X as the batchsize, so the multiplication will return a wrong result.
                                                                                            #Thats why we add an extra dimension, and transpose the vector. At the end we will have a vector with shape [batchsize,1,784].
                                                                                            #And the multiplication result will be correct.
                                                                                            
				        Z = tf.matmul(Xaux, memW) 	#Matrix multiplication between input and mask. With output shape [batchsize,1,128]
				        Z = tf.reshape(Z,[self.batch_size,tf.shape(Z)[-1]]) #We need to reshape again because we are working with column vectors. The output shape must be[batchsize,128]
				        Z = tf.add(Z,membias) #FInally, we add the bias error mask 
				        #Z = self.forward(self.W,self.bias,self.Xaux)
				else: #if we have pool attribute the layer will train with a pool of error matrices created during the forward propagation.
				    if(self.Wstd !=0):							         
				        Werr = abs(1+tf.random.normal(shape=[self.pool,tf.cast(row,tf.int32),self.output_size],stddev=self.Wstd,dtype=self.d_type))#tf.squeeze(Werr, axis=0)				#This is necessary because gather add an extra dimension. Squeeze remove this dimension.			 
                                                                        #That means, with a weights shape of [784,128] and a batch size of 256. Werr should be a tensor with shape	
                                                                        #[256,784,128], but gather return us a tensor with shape [1,256,784,128], so we remove that 1 with squeeze.
				    else:
				        Werr = self.Werr
        
				    if(self.Bstd !=0):								#For the bias is exactly the same situation
				        Berr = abs(1+tf.random.normal(shape=[self.pool,self.output_size],stddev=self.Bstd,dtype=self.d_type))#tf.squeeze(Berr,axis=0)
				    else:
				        Berr = self.Berr	
                        
				    newBatch = tf.cast(tf.floor(tf.cast(self.batch_size/self.pool,dtype=tf.float16)),dtype=tf.int32)
				    Z = tf.matmul(self.X[0:newBatch], weights*Werr[0]) 	#Matrix multiplication between input and mask. With output shape [batchsize,1,128]
				    Z = tf.reshape(Z,[newBatch,tf.shape(Z)[-1]]) #We need to reshape again because we are working with column vectors. The output shape must be[batchsize,128]
				    Z = tf.add(Z,self.bias*Berr[0]) #FInally, we add the bias error mask 
				    for i in range(self.pool-1):
				        Z1 = tf.matmul(self.X[(i+1)*newBatch:(i+2)*newBatch], weights*Werr[i+1])
				        Z1 = tf.add(Z1,self.bias*Berr[i+1])                                           
				        Z = tf.concat([Z,Z1],axis=0)                                                 

			else:
				if(self.isBin=='yes'):
					self.memW = self.sign(self.W)*self.Werr 
				else:
					self.memW = self.W*self.Werr
				bias = self.bias*self.Berr
				Z = tf.add(tf.matmul(self.X,self.memW),bias) #Custom FC layer operation when we don't have Wstd or Bstd.

		else:
		    #This part of the code will be executed during the inference
			if(self.Wstd != 0 or self.Bstd !=0):
				if(self.Wstd !=0):
					Werr = self.infWerr
				else:
					Werr = self.Werr
				if(self.Bstd != 0):
					Berr = self.infBerr
				else:
					Berr = self.Berr
			else:
				Werr = self.Werr
				Berr = self.Berr
			if(self.isBin=='yes'):
				weights =tf.math.sign(self.W)*Werr
			else:
				weights = self.W*Werr
			bias = self.bias*Berr		
			Z = tf.add(tf.matmul(self.X, weights), bias)
					
		return Z
	def slice_batch(self,miniBatch,N,row):        
		if(self.Wstd != 0):
			Werr = abs(1+tf.random.normal(shape=[miniBatch,tf.cast(row,tf.int32),self.output_size],stddev=self.Wstd,dtype=self.d_type))
		else:
			Werr = self.Werr 
		if(self.Bstd != 0):
			Berr =  abs(1+tf.random.normal(shape=[miniBatch,self.output_size],stddev=self.Bstd,dtype=self.d_type))
		else:
			Berr = self.Berr 
		memW = weights*Werr
		membias = self.bias*Berr                                    
		Xaux = tf.reshape(self.X[N*miniBatch:(N+1)*miniBatch], [miniBatch,1,row]) 
																					
		Z = tf.matmul(Xaux, memW) 	#Matrix multiplication between input and mask. With output shape [batchsize,1,128]         
		Z = tf.reshape(Z,[miniBatch,tf.shape(Z)[-1]]) #We need to reshape again because we are working with column vectors. The output shape must be[batchsize,128]           
		Z = tf.add(Z,membias) #FInally, we add the bias error mask                       
		#Z = self.forward(self.W,self.bias,Xaux)
		#tf.print('Z dims: ',tf.shape(Z))                             
		return Z  

	#THis is only for saving purposes. Does not affect the layer performance.	
	def get_config(self):
		config = super(FC_AConnect, self).get_config()
		config.update({
			'output_size': self.output_size,
			'Wstd': self.Wstd,
			'Bstd': self.Bstd,
			'isBin': self.isBin,
            'pool' : self.pool,
            'Slice': self.Slice,
            'd_type': self.d_type,
            'weights_regularizer': self.weights_regularizer,
            'bias_regularizer' : self.bias_regularizer})
		return config

	def validate_init(self):
		if self.output_size <= 0:
		    raise ValueError('Unable to build a Dense layer with 0 or negative dimension. ' 'Output size: %d' %(self.output_size,))
		if self.Wstd > 1 or self.Wstd < 0:
		    raise ValueError('Wstd must be a number between 0 and 1. \n' 'Found %d' %(self.Wstd,)) 
		if self.Bstd > 1 or self.Bstd < 0:
		    raise ValueError('Bstd must be a number between 0 and 1. \n' 'Found %d' %(self.Bstd,))    
		if not isinstance(self.isBin, str):
		    raise TypeError('isBin must be a string, yes or no. ' 'Found %s' %(type(self.isBin),))
		if self.pool is not None and not isinstance(self.pool, int):
		    raise TypeError('pool must be a integer. ' 'Found %s' %(type(self.pool),))                                                                      		    
		
	@tf.custom_gradient
	def sign(self,x):
		y = tf.math.sign(x)
		def grad(dy):
			dydx = tf.divide(dy,abs(x)+1e-5)
			return dydx
		return y, grad
###HOW TO IMPLEMENT MANUALLY THE BACKPROPAGATION###		
"""	
	@tf.custom_gradient
	def forward(self,W,bias,X):
		if(self.Wstd != 0):
			mWerr = abs(1+tf.random.normal(shape=[self.miniBatch,tf.cast(self.row,tf.int32),self.output_size],stddev=self.Wstd,dtype=self.d_type))
		else:
			mWerr = self.Werr 
		if(self.Bstd != 0):
			Berr =  abs(1+tf.random.normal(shape=[self.miniBatch,self.output_size],stddev=self.Bstd,dtype=self.d_type))
		else:
			Berr = self.Berr   
		if(self.isBin=='yes'):
			weights = tf.math.sign(W)			#Binarize the weights
		else:
			weights = W
		weights = tf.expand_dims(weights, axis=0)
		loc_W = weights*mWerr				#Get the weights with the error matrix included. Also takes the binarization error when isBin=yes
		bias = tf.expand_dims(bias, axis=0)
		loc_bias = bias*Berr
		Z = tf.matmul(X,loc_W)
		Z = tf.reshape(Z, [self.miniBatch,tf.shape(Z)[-1]]) #Reshape Z to column vector
		Z = tf.add(Z, loc_bias) # Add the bias error mask
		def grad(dy):
			if(self.isBin=="yes"):
				layerW = tf.expand_dims(W, axis=0)
				Werr = loc_W/layerW		#If the layer is binary we use Werr as W*/layer.W as algorithm 3 describes.
			else:
				Werr = mWerr  #If not, Werr will be the same matrices that we multiplied before
			dy = tf.reshape(dy, [self.miniBatch,1,tf.shape(dy)[-1]]) #Reshape dy to [batch,1,outputsize]
			dX = tf.matmul(dy,loc_W, transpose_b=True) #Activations gradient
			dX = tf.reshape(dX, [self.miniBatch, tf.shape(dX)[-1]])
			dWerr = tf.matmul(X,dy,transpose_a=True) #Gradient for the error mask of weights
			dBerr = tf.reshape(dy, [self.miniBatch,tf.shape(dy)[-1] ]) #Get the gradient of the error mask of bias with property shape
			dW = dWerr*Werr #Get the property weights gradient
			dW = tf.reduce_sum(dW, axis=0)
			dB = dBerr*Berr #Get the property bias gradient
			dB = tf.reduce_sum(dB, axis=0)
			return dW,dB,dX
		return Z, grad """
###########################################################################################################3
"""
Convolutional layer with A-Connect
INPUT ARGUMENTS:
-filters: Number of filter that you want to use during the convolution.(Also known as output channels)
-kernel_size: List with the dimension of the filter. e.g. [3,3]. It must be less than the input data size
-Wstd and Bstd: Weights and bias standard deviation for training
-pool: Number of error matrices that you want to use.
-isBin: string yes or no, whenever you want binary weights
-strides: Number of strides (or steps) that the filter moves during the convolution
-padding: "SAME" or "VALID". If you want to keep the same size or reduce it.
-Op: 1 or 2. Which way to do the convolution you want to use. The first option is slower but has less memory cosumption and the second one is faster
but consumes a lot of memory.
-Slice: Optional parameter. Used to divide the batch into 2,4 or 8 minibatches of size batch/N.
-d_type: Type of the parameters that the layers will create. Supports fp16, fp32 and fp64 
"""
class Conv_AConnect(tf.keras.layers.Layer):
	def __init__(self,filters,kernel_size,strides=1,padding="VALID",Wstd=0,Bstd=0,pool=None,isBin='no',Op=1
    ,Slice=1,d_type=tf.dtypes.float32,weights_regularizer=None,bias_regularizer=None,**kwargs):
		super(Conv_AConnect, self).__init__()
		self.filters = filters
		self.kernel_size = kernel_size
		self.Wstd = Wstd
		self.Bstd = Bstd
		self.pool = pool        
		self.isBin = isBin
		self.strides = strides
		self.padding = padding
		self.Op = Op
		self.Slice = Slice
		self.d_type = d_type
		self.weights_regularizer = tf.keras.regularizers.get(weights_regularizer)                  #Weights regularizer. Default is None
		self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)                        #Bias regularizer. Default is None        
		self.validate_init()        
	def build(self,input_shape):
		self.shape = list(self.kernel_size) + list((int(input_shape[-1]),self.filters)) ### Compute the shape of the weights. Input shape could be [batchSize,H,W,Ch] RGB

		self.W = self.add_weight('kernel',
								  shape = self.shape,
								  initializer = "glorot_uniform",
                                  dtype=self.d_type,
                                  regularizer = self.weights_regularizer,
								  trainable=True)				  
		self.bias = self.add_weight('bias',
									shape=(self.filters,),
									initializer = 'zeros',
                                    dtype=self.d_type,
                                    regularizer = self.bias_regularizer,
									trainable=True)
		if(self.Wstd != 0 or self.Bstd != 0): #If the layer will take into account the standard deviation of the weights or the std of the bias or both
			if(self.Bstd != 0):
				self.infBerr = abs(1+tf.random.normal(shape=[self.filters,],stddev=self.Bstd)) #Bias error vector for inference
				self.infBerr = self.infBerr.numpy()  #It is necessary to convert the tensor to a numpy array, because tensors are constant and therefore cannot be changed
													 #This was necessary to change the error matrix/array when Monte Carlo was running.
																	
			else:
				self.Berr = tf.constant(1,dtype=self.d_type)
			if(self.Wstd !=0): 
				self.infWerr = abs(1+tf.random.normal(shape=self.shape,stddev=self.Wstd)) #Weight matrix for inference
				self.infWerr = self.infWerr.numpy()										 
				 
			else:
				self.Werr = tf.constant(1,dtype=self.d_type)
		else:
			self.Werr = tf.constant(1,dtype=self.d_type) #We need to define the number 1 as a float32.
			self.Berr = tf.constant(1,dtype=self.d_type)
		super(Conv_AConnect, self).build(input_shape)
	def call(self,X,training):
		self.X = tf.cast(X, dtype=self.d_type)       
		self.batch_size = tf.shape(self.X)[0]
		if(training):
			if(self.Wstd != 0 or self.Bstd != 0):
				if(self.isBin=='yes'):
				    weights=self.sign(self.W)
				else:
				    weights=self.W				
				if(self.pool is None):				
				    if(self.Op == 1):
				        if(self.Slice == 2): #Slice the batch into 2 minibatches of size batch/2
				            miniBatch = tf.cast(self.batch_size/2,dtype=tf.int32)                                                                  
				            Z1 = self.slice_batch(weights,miniBatch,0,self.strides) #Takes a portion from 0:minibatch
				            Z2 = self.slice_batch(weights,miniBatch,1,self.strides) #Takes a portion from minibatch:2*minibatch
				            Z = tf.concat([Z1,Z2],axis=0)
				        elif(self.Slice == 4):
					        miniBatch = tf.cast(self.batch_size/4,dtype=tf.int32) #Slice the batch into 4 minibatches of size batch/4
					        Z = self.slice_batch(weights,miniBatch,0,self.strides) #Takes a portion from 0:minibatch
					        for i in range(3):
					            Z1 = self.slice_batch(weights,miniBatch,i+1,self.strides) #Takes a portion from i*minibatch:(i+1)*minibatch
					            Z = tf.concat([Z,Z1],axis=0)                                                                                                                                                                        
				        elif(self.Slice == 8):
					        miniBatch = tf.cast(self.batch_size/8,dtype=tf.int32) #Slice the batch into 8 minibatches of size batch/8
					        Z = self.slice_batch(weights,miniBatch,0,self.strides) #Takes a portion from 0:minibatch                  
					        for i in range(7):
					            Z1 = self.slice_batch(weights,miniBatch,i+1,self.strides) #Takes a portion from i*minibatch:(i+1)*minibatch
					            Z = tf.concat([Z,Z1],axis=0)                          
				        else:
				            if(self.Wstd != 0):
				                Werr = abs(1+tf.random.normal(shape=list((self.batch_size,))+self.shape,stddev=self.Wstd,dtype=self.d_type))#tf.squeeze(Werr, axis=0)
				            else:
				                Werr = self.Werr
				            weights = tf.expand_dims(weights,axis=0)
				            memW = tf.multiply(weights,Werr)
				            if(self.Bstd != 0):
				                Berr =  abs(1+tf.random.normal(shape=[self.batch_size,self.filters],stddev=self.Bstd,dtype=self.d_type))#tf.squeeze(Berr, axis=0)
				            else:
					            Berr = self.Berr
				            bias = tf.expand_dims(self.bias,axis=0)
				            membias = tf.multiply(bias,Berr)
				            membias = tf.reshape(membias,[self.batch_size,1,1,tf.shape(membias)[-1]])                    
				            Z = tf.squeeze(tf.map_fn(self.conv,(tf.expand_dims(self.X,1),memW)
                            ,fn_output_signature=self.d_type),axis=1)#tf.nn.convolution(Xaux,memW,self.strides,self.padding)
				            Z = tf.reshape(Z, [self.batch_size, tf.shape(Z)[1],tf.shape(Z)[2],tf.shape(Z)[3]])
				            Z = Z+membias
				##################################################################################################################################
				    else:
				        strides = [1,self.strides,self.strides,1]
				        if(self.Slice == 2): #Slice the batch into 2 minibatches of size batch/2
				            miniBatch = tf.cast(self.batch_size/2,dtype=tf.int32)                                                                  
				            Z1 = self.slice_batch(weights,miniBatch,0,strides) #Takes a portion from 0:minibatch
				            Z2 = self.slice_batch(weights,miniBatch,1,strides) #Takes a portion from minibatch:2*minibatch
				            Z = tf.concat([Z1,Z2],axis=0)
				        elif(self.Slice == 4):
					        miniBatch = tf.cast(self.batch_size/4,dtype=tf.int32) #Slice the batch into 4 minibatches of size batch/4
					        Z = self.slice_batch(weights,miniBatch,0,strides) #Takes a portion from 0:minibatch
					        for i in range(3):
					            Z1 = self.slice_batch(weights,miniBatch,i+1,strides) #Takes a portion from i*minibatch:(i+1)*minibatch
					            Z = tf.concat([Z,Z1],axis=0)                                                                                                                                                                        
				        elif(self.Slice == 8):
					        miniBatch = tf.cast(self.batch_size/8,dtype=tf.int32) #Slice the batch into 8 minibatches of size batch/8
					        Z = self.slice_batch(weights,miniBatch,0,strides) #Takes a portion from 0:minibatch                  
					        for i in range(7):
					            Z1 = self.slice_batch(weights,miniBatch,i+1,strides) #Takes a portion from i*minibatch:(i+1)*minibatch
					            Z = tf.concat([Z,Z1],axis=0)                                                                                                                                                              
				        else:     
					        if(self.Wstd != 0):
					            Werr = abs(1+tf.random.normal(shape=list((self.batch_size,))+self.shape,stddev=self.Wstd,dtype=self.d_type))#tf.squeeze(Werr, axis=0)
					        else:
					            Werr = self.Werr
					        weights = tf.expand_dims(weights,axis=0)
					        memW = tf.multiply(weights,Werr)
					        if(self.Bstd != 0):
					            Berr =  abs(1+tf.random.normal(shape=[self.batch_size,self.filters],stddev=self.Bstd,dtype=self.d_type))#tf.squeeze(Berr, axis=0)
					        else:
					            Berr = self.Berr
					        bias = tf.expand_dims(self.bias,axis=0)
					        membias = tf.multiply(bias,Berr)
					        membias = tf.reshape(membias,[self.batch_size,1,1,tf.shape(membias)[-1]])                                                                                                                                               
					        inp_r, F = reshape(self.X,memW) #Makes the reshape from [batch,H,W,ch] to [1,H,W,Ch*batch] for input. For filters from [batch,fh,fw,Ch,out_ch]  to
                                                                        #[fh,fw,ch*batch,out_ch]
					        Z = tf.nn.depthwise_conv2d(
                                    inp_r,
                                    filter=F,
                                    strides=strides,
                                    padding=self.padding)
					        Z = Z_reshape(Z,memW,self.X,self.padding,self.strides) #Output shape from convolution is [1,newH,newW,batch*Ch*out_ch] so it is reshaped to [newH,newW,batch,Ch,out_ch]
                                                                #Where newH and newW are the new image dimensions. This depends on the value of padding
                                                                #Padding same: newH = H  and newW = W
                                                                #Padding valid: newH = H-fh+1 and newW = W-fw+1
					        Z = tf.transpose(Z, [2, 0, 1, 3, 4]) #Get the property output shape [batch,nH,nW,Ch,out_ch]
					        Z = tf.reduce_sum(Z, axis=3)		#Removes the input channel dimension by adding all this elements                                                        
					        Z = membias+Z					#Add the bias
				else:
				    if(self.Wstd != 0):
				        Werr = abs(1+tf.random.normal(shape=list((self.pool,))+self.shape,stddev=self.Wstd,dtype=self.d_type))#tf.squeeze(Werr, axis=0)
				    else:
				        Werr = self.Werr      

				    if(self.Bstd != 0):

				        Berr =  abs(1+tf.random.normal(shape=[self.pool,self.filters],stddev=self.Bstd,dtype=self.d_type))#tf.squeeze(Berr, axis=0)
				    else:
				        Berr = self.Berr

				    newBatch = tf.cast(tf.floor(tf.cast(self.batch_size/self.pool,dtype=tf.float16)),dtype=tf.int32)
				    Z = tf.nn.conv2d(self.X[0:newBatch], weights*Werr[0],strides=[1,self.strides,self.strides,1],padding=self.padding) 
				    Z = tf.add(Z,self.bias*Berr[0]) #FInally, we add the bias error mask 
				    for i in range(self.pool-1):
				        Z1 = tf.nn.conv2d(self.X[(i+1)*newBatch:(i+2)*newBatch], weights*Werr[i+1],strides=[1,self.strides,self.strides,1],padding=self.padding)                    
				        Z1 = tf.add(Z1,self.bias*Berr[i+1])                                           
				        Z = tf.concat([Z,Z1],axis=0)                                                   
			else:
				if(self.isBin=='yes'):
					weights=self.sign(self.W)*self.Werr
				else:
					weights=self.W*self.Werr
				Z = self.bias*self.Berr+tf.nn.conv2d(self.X,weights,self.strides,self.padding)				
		else:
			if(self.Wstd != 0 or self.Bstd !=0):
				if(self.Wstd !=0):
					Werr = self.infWerr
				else:
					Werr = self.Werr
				if(self.Bstd != 0):
					Berr = self.infBerr
				else:
					Berr = self.Berr
			else:
				Werr = self.Werr
				Berr = self.Berr		
			if(self.isBin=='yes'):
				weights=tf.math.sign(self.W)*Werr
			else:
				weights=self.W*Werr	
			bias = self.bias*Berr                
			Z = bias+tf.nn.conv2d(self.X,weights,self.strides,self.padding)	           								
		return Z
	def slice_batch(self,weights,miniBatch,N,strides):
		if(self.Wstd != 0):
			Werr = abs(1+tf.random.normal(shape=list((miniBatch,))+self.shape,stddev=self.Wstd,dtype=self.d_type))#tf.squeeze(Werr, axis=0)
		else:
			Werr = self.Werr

		weights = tf.expand_dims(weights,axis=0)
		memW = tf.multiply(weights,Werr)
		if(self.Bstd != 0):
			Berr =  abs(1+tf.random.normal(shape=[miniBatch,self.filters],stddev=self.Bstd,dtype=self.d_type))#tf.squeeze(Berr, axis=0)
		else:
			Berr = self.Berr
		bias = tf.expand_dims(self.bias,axis=0)
		membias = tf.multiply(bias,Berr)
		membias = tf.reshape(membias,[miniBatch,1,1,tf.shape(membias)[-1]])                    
		if self.Op == 1:
			Z = tf.squeeze(tf.map_fn(self.conv,(tf.expand_dims(self.X[N*miniBatch:(N+1)*miniBatch],1),memW),
            fn_output_signature=self.d_type),axis=1)
			Z = tf.reshape(Z, [miniBatch, tf.shape(Z)[1],tf.shape(Z)[2],tf.shape(Z)[3]])
			Z = Z+membias 
		else:                       
		    inp_r, memW = reshape(self.X[N*miniBatch:(N+1)*miniBatch],memW)
		    Z = tf.nn.depthwise_conv2d(
                                    inp_r,
                                    filter=memW,
                                    strides=strides,
                                    padding=self.padding)
		    Z = Z_reshape(Z,Werr,self.X[N*miniBatch:(N+1)*miniBatch],self.padding,self.strides)
		    Z = tf.transpose(Z, [2, 0, 1, 3, 4])
		    Z = tf.reduce_sum(Z, axis=3)
		    Z = Z+membias  
		#Z = tf.cond(tf.less_equal(inp_shape[1],100),cond2(miniBatch,N,memW,Werr,membias),cond1(miniBatch,N,memW,membias))                    
		return Z   

	def conv(self,tupla):
		x,kernel = tupla
		return tf.nn.convolution(x,kernel,strides=self.strides,padding=self.padding) 
	def validate_init(self):
		if not isinstance(self.filters, int):
		    raise TypeError('filters must be an integer. ' 'Found %s' %(type(self.filters),))
		if self.Wstd > 1 or self.Wstd < 0:
		    raise ValueError('Wstd must be a number between 0 and 1. \n' 'Found %d' %(self.Wstd,)) 
		if self.Bstd > 1 or self.Bstd < 0:
		    raise ValueError('Bstd must be a number between 0 and 1. \n' 'Found %d' %(self.Bstd,))    
		if not isinstance(self.isBin, str):
		    raise TypeError('isBin must be a string, yes or no. ' 'Found %s' %(type(self.isBin),))
		if self.pool is not None and not isinstance(self.pool, int):
		    raise TypeError('pool must be a integer. ' 'Found %s' %(type(self.pool),))             
	def get_config(self):
		config = super(Conv_AConnect, self).get_config()
		config.update({
			'filters': self.filters,
			'kernel_size': self.kernel_size,
			'Wstd': self.Wstd,
			'Bstd': self.Bstd,
            'pool': self.pool,
			'isBin': self.isBin,
			'strides': self.strides,
			'padding': self.padding,
            'Op': self.Op,
			'Slice': self.Slice,
			'd_type': self.d_type})
		return config
	@tf.custom_gradient
	def sign(self,x):
		y = tf.math.sign(x)
		def grad(dy):
			dydx = tf.divide(dy,abs(x)+1e-5)
			return dydx
		return y, grad		
############################AUXILIAR FUNCTIONS##################################################
def reshape(X,F): #Used to reshape the input data and the noisy filters
    batch_size=tf.shape(X)[0]    
    H = tf.shape(X)[1]
    W = tf.shape(X)[2]
    channels_img = tf.shape(X)[3]
    channels = channels_img
    fh = tf.shape(F)[1]
    fw = tf.shape(F)[2]    
    out_channels = tf.shape(F)[-1]
    F = tf.transpose(F, [1, 2, 0, 3, 4])
    F = tf.reshape(F, [fh, fw, channels*batch_size, out_channels]) 
    inp_r = tf.transpose(X, [1, 2, 0, 3])
    inp_r = tf.reshape(inp_r, [1, H, W, batch_size*channels_img])
    return inp_r, F          
def Z_reshape(Z,F,X,padding,strides): #Used to reshape the output of the layer
    batch_size=tf.shape(X)[0]
    H = tf.shape(X)[1]
    W = tf.shape(X)[2]    
    channels_img = tf.shape(X)[3]
    channels = channels_img
    fh = tf.shape(F)[1]
    fw = tf.shape(F)[2]    
    out_channels = tf.shape(F)[-1]
    #tf.print(fh)    
    if padding == "SAME":
        return tf.reshape(Z, [tf.floor(tf.cast((H)/strides,dtype=tf.float16)), tf.floor(tf.cast((W)/strides,dtype=tf.float16)), batch_size, channels, out_channels])
    if padding == "VALID":
        return tf.reshape(Z, [tf.floor(tf.cast((H-fh)/strides,dtype=tf.float16))+1, tf.floor(tf.cast((W-fw)/strides,dtype=tf.float16))+1, batch_size, channels, out_channels])
    #return out              		
			
